import asyncio
import json
import logging

import aioamqp

from .app import Component


class AMQPServer(Component):
    CONNECTION_DELAY = 1

    def __init__(self, app, config, handler):
        super(AMQPServer, self).__init__()
        self.app = app
        self.loop = app.loop
        self.config = {
            'host': config.get('host', 'localhost'),
            'port': config.get('port', ),
            'virtualhost': config.get('virtualhost', '/'),
            'login': config.get('login', 'guest'),
            'password': config.get('password', 'guest'),
            'durable': config.get('durable', True),
        }
        self.exchange_name = config.get('exchange_name', 'mocker')
        self.exchange_type = config.get('exchange_type', 'topic')
        self.exchange_durable = config.get('exchange_durable', True)
        self.transport = None
        self.protocol = None
        self.channel = None
        self.consumer = handler(self)
        self.consumers_list = []
        self.started = []

    async def prepare(self):
        self.app.log_info("Preparing to start AMQP server...")
        self.consumers_list = self.consumer.routes

    async def start(self):
        while True:
            try:
                await self.connect()
                await self.channel.exchange_declare(
                    exchange_name=self.exchange_name,
                    type_name=self.exchange_type,
                    durable=self.exchange_durable,
                )
                for route in self.consumers_list:
                    # name, queue_name, func
                    await self.channel.queue_declare(
                        queue_name=route['queue_name'],
                        #durable=self.config['durable'],
                    )
                    await self.channel.queue_bind(          # TODO: timeout
                        exchange_name=self.exchange_name,
                        queue_name=route['queue_name'],
                        routing_key=route['routing_key']
                    )
                    await self.channel.basic_consume(       # TODO: get tag
                        route['handler'],
                        queue_name=route['queue_name']
                    )
                    self.app.log_info("Start consumer '%s'." % route['route_name'])
            except Exception as e:
                await asyncio.sleep(5)
                self.app.log_err('AMQP connect error: %s %s' % (self.__class__.__name__, str(e)))
                self.app.log_err('DSN: %s' % json.dumps(self.config))
                self.app.log_err('Trying to reconnect to rabbit...')
            else:
                break

    async def stop(self):
        # TODO: сделать через basic_cancel и тег
        # await self.channel.stop_consuming()
        await self.disconnect()
        self.app.log_info("Stop AMQP server.")

    # async def heartbeat(self, protocol):
    #     while True:
    #         await asyncio.sleep(30)  # issue manual heartbeat every 20 seconds
    #         await protocol.send_heartbeat()

    async def connect(self):
        try:
            self.transport, self.protocol = await aioamqp.connect(
                **self.config,
                loop=self.loop,
                on_error=self.error_callback,
                heartbeat=30
            )
            self.channel = await self.protocol.channel()
            # asyncio.ensure_future(self.heartbeat(self.protocol))
        except aioamqp.AmqpClosedConnection as e:
            self.app.log_info('[x] Connection error %s' % str(e))
            await asyncio.sleep(self.CONNECTION_DELAY)
            await self.connect()

    async def disconnect(self):
        if self.transport and self.protocol:
            await self.protocol.close()
            self.transport.close()

    async def error_callback(self, error):
        if isinstance(error, aioamqp.AmqpClosedConnection):
            await asyncio.sleep(self.CONNECTION_DELAY)
            self.app.log_info("Restarting AMQP server.")
            await self.start()



# -*- coding: utf-8 -*-

"""Интерфейс trac_uni."""


try:
    import config.tracuni_async_config
    async_variant = getattr(config.tracuni_async_config, "ASYNC_VARIANT", "ASYNCIO")
    suppress_bd_spans = getattr(config.tracuni_async_config, "SUPPRESS_BD_SPANS", False)
except ImportError:
    async_variant = "ASYNCIO"
    suppress_bd_spans = False
import tracuni.misc.select_coroutine as select_coroutine
select_coroutine.ASYNC_VARIANT = async_variant
import tracuni.schema.engine.engine as engine
engine.SUPPRESS_BD_SPANS = suppress_bd_spans

from .intercept import (
    tracer_point,
)
from .adapter import fab_tracer
from .adapter_tornado import init_tornado_loop
from .side_steps.db_logger.logger import Logger as DBLogger
from .schema.pipe_methods import (
    pipe_cut,
    pipe_catch_essential,
)

from .define import errors

from .define.type import (
    Schema,
    Extractors,
    RuleSet,
    Rule,
    Origins,
    Origin,
    OriginSection,
    Destination,
    DestinationSection,
    Pipeline,
    PipelineTee,
    TeeDescriptor,
    PipeCommand,
    Stage,
    Variant,
    SpanSide,
    APIKind,
    Part,
)

__author__ = """Konstantin Gonciarou"""
__email__ = 'konstantin.goncharov@inplat.ru'
__version__ = '0.4.11'

__all__ = (
    # interceptor
    'fab_tracer',  # инициализация адаптера для работы с aiohttp приложением
    'init_tornado_loop',  # инициализация для работы с Tornado приложением
    'tracer_point',  # декоратор перехвата вызова точки и запуска извлечения
    # Types
    'Schema',  # Списки правил по всем вариантам
    'Extractors',  # Списки правил конкретного варианта по этапам
    'RuleSet',  # Список правил для обработки конкретной точки
    'Rule',  # Правило - откуда, куда. на каком этапе переместить данные и
             # как их обработать, либо команда управления применением правил
    'Stage',  # этап обработки одной точки
    'Origins',  # Список описаний источников данных
    'Origin',  # Описание одного источника данных
    'OriginSection',  # Разделы извлечения данных
    'Pipeline',  # Список методов последовательной обработки данных
    'PipeCommand',  # Обработчик-команда упрвления самим процессом обработки
    'PipelineTee',  # Разветвление потока данных при обработке
    'TeeDescriptor',  # Описание работы разветвления
    'Destination',  # Описание назначения обработанных данных
    'DestinationSection',  # Разделы, куда попадают обработанные данные
    'Variant',  # Вариант обработки
    'SpanSide',  # Сторона варианта
    'APIKind',  # Вид интерфейса варианта
    'Part',
    # auxiliary logging
    'DBLogger',
    'errors',
    'pipe_cut',
    'pipe_catch_essential',
)

"""Структура модулей по смысловым блокам

 ☐ Сопряжение:
        adapter
        misc.keep_tornado
   ☐ Инициализация и конфигурация при старте приложения
   ☐ Выбор асинхронного механизма
   ☐ Сохранение родительского контекста участка

 ☐ Перехват:
        intercept
        point_accessor
   ☐
   ☐ Ссылки на данные

 ☐ Управление потоком:
        schema.engine ( cache | feed | engine )
        schema.standard
   ☐ Составление схем
   ☐ Проведение данных через обработку

 ☐ Фиксация:
        span
        side_steps ( db_logger | stat )
   ☐ Создание участков
   ☐ Наполнение участков
   ☐ Параллельная запись в другие сервисы

 ☐ Обработка:
        schema.builtin ( ext_* | pipe | helper )
   ☐ Подготовка данных для записи

 ☐ Вспомогательные:
        define ( type | const | errors )
        misc.helper
   ☐ Общие вспомогательные методы
   ☐ Структуры данных
   ☐ Внутренние настроечные данные
   ☐ Пользовательские исключения
"""



import json
import asyncio
import logging
from datetime import datetime

import pytz
import aioamqp

from tracuni.misc.helper import dict_from_json
from tracuni.schema.pipe_methods import pipe_mask_secret_catch_essentials


TIME_FORMAT = "%Y-%m-%d %H:%M:%S%z"


class Logger(object):
    def __init__(self):
        self.config = json.load(open('config/main.json', encoding='utf8'))

    def __call__(self, func):
        async def wrapper(that, *args, **kwargs):
            stamp_begin = datetime.now(pytz.utc).strftime(TIME_FORMAT)
            data = await func(that, *args, **kwargs)
            stamp_end = datetime.now(pytz.utc).strftime(TIME_FORMAT)
            body = kwargs.get('body')
            if body:
                body = dict_from_json(body)
            else:
                body = args[0]
            response = kwargs.get('response')
            if response:
                response = dict_from_json(response)
            else:
                response = {}
            if data:
                log_data = {
                    'stamp_begin': stamp_begin,
                    'stamp_end': stamp_end,
                    'func_name': func.__name__,
                    'url': args[-1],
                    'request_headers': kwargs.get('headers', {}),
                    'request': body,
                    'response_headers': data.get('response_headers', {}),
                    'response': response,
                    'service_name': self.config['logging'].get(
                        'service_name',
                        self.config['logging'].get('tracer_svc_name', '')
                    ),
                    'external_key': data.get('external_key', '<unknown>'),
                    'http_code': data.get('http_code', 200)
                }
                await self.log(log_data)
            return data
        return wrapper

    async def log(self, log_data):
        # Connecting to rabbit_mq
        time_waited = 0
        to_wait = 1.5
        while True:
            try:
                transport, protocol = await aioamqp.connect(
                    **self.config['logging']
                )
                break
            except (OSError, aioamqp.AmqpClosedConnection) as e:
                to_wait = round(min(30, (to_wait ** 1.5)), 2)
                logging.info(
                    "[x] Failed to connect to RabbitMQ: "
                    "%s. Waiting %s seconds...",
                    e,
                    to_wait
                )
                await asyncio.sleep(to_wait)
                time_waited += to_wait

        channel = await protocol.channel()
        await channel.publish(
            pipe_mask_secret_catch_essentials(json.dumps(log_data))[0],
            self.config['logging']['exchange_name'],
            self.config['logging']['routing_key']
        )
        await protocol.close()
        # ensure the socket is closed.
        transport.close()
        logging.info('[x] Log sent')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Перехват вызова интересующей точки
    Во время перехвата запускается процесс сбора, обработки и записи данных,
    согласно связаным с перехватчиком набором правил
"""
import inspect
import logging
import traceback
from functools import wraps
import sys

from tracuni.adapter import TracerAdapter
from tracuni.schema.engine.engine import SchemaEngine
from tracuni.define.type import (
    Variant,
    RuleSet,
    Stage,
    Part,
)
from tracuni.misc.helper import (
    PostgreSQLError,
    log_tracer_error,
)
from tracuni.define.errors import TracerCollectedException
from tracuni.point.accessor import PointAccessor
from tracuni.misc.select_coroutine import get_coroutine_decorator
async_decorator = get_coroutine_decorator()


def tracer_point(
    variant: Variant,
    remote_name: str = None,
    point_ruleset: RuleSet = None,
    part: Part = None,
    reraise_exceptions: bool = True,
    pass_itself_as_kwargs_item: bool = False,
):
    """

    Parameters
    ----------
    reraise_exceptions
    variant
        Назначить вариант обработки точки
    remote_name
        Здесь можно передать пользовательское имя удаленной точки
        при этом можно использовать и стандартное, указав в строке место
        подстановки с именем standard: 'mine: {standard}'
    point_ruleset
        Здесь можно передать свой вариант настройки извлечения формат такой
        же как значение одного из ключей стандартной настройки она заменит
        стандартную настройку конкретного варианта для данной точки,
        если же последним элементом пользовательской настройки будет строка
        '...', то пользовательская объединится со стандартной с приоритетом
        пользовательской в любом случае далее удут применены менее
        специфичные и менее приоритетные настройки, относящиеся к данной точке
    part
        разрывает обработку на две части, закрытие участка происходит во второй

    Returns
    -------
        Декорированный метод, вызов которого будет трассироваться

    """
    # создание экземпляра обработчика
    schema_engine = SchemaEngine(variant, point_ruleset)
    # подписываемся, чтобы знать состояние трассера
    TracerAdapter.register_state_observer(schema_engine)

    def wrapper(point_coro):

        @wraps(point_coro)
        @async_decorator
        def decorated(*args, **kwargs):
            result = {}
            point = None
            debug_is_on = False

            if schema_engine.is_tracer_disabled():
                # сразу запускаем обернутую точку, если трассер выключен,
                # не соединился или не создан
                result = yield from point_coro(*args, **kwargs)
                return result

            if part in (Part.CLOSE, Part.CLOSE_AFTER):
                open_point = None
                open_span_id = None
                try:
                    if part == Part.CLOSE_AFTER:
                        result = yield from point_coro(*args, **kwargs)
                    open_span_context = schema_engine.adapter.keep_context.get('span')
                    if open_span_context:
                        open_span_id = open_span_context.get('X-B3-SpanId')
                        if open_span_id:
                            open_point = schema_engine.adapter.open_span_points.get(open_span_id)
                            if isinstance(result, TracerCollectedException):
                                open_point.register_error(result.exc, (type(result.exc), result.exc,
                                                              result.exc.__traceback__))
                                result = result.result
                        yield from open_point.register_stage(Stage.POST)
                except Exception as exc:
                    if open_point:
                        if open_span_id:
                            schema_engine.adapter.open_span_points.pop(open_span_id, None)
                        open_point.register_own_error(exc)
                finally:
                    if open_span_id:
                        schema_engine.adapter.open_span_points.pop(open_span_id, None)
                    if part == Part.CLOSE:
                        result = yield from point_coro(*args, **kwargs)
                    return result

            try:

                debug_is_on = schema_engine.adapter.config.debug

                point = PointAccessor.fab_default(
                    args,
                    kwargs,
                    point_coro,
                    schema_engine,
                    remote_name,
                    pass_itself_as_kwargs_item,
                )
                point.stack_frames_to_context(inspect.currentframe())

                # создаем спан сконфигрированный по варианту и связанный с
                # ним и точкой
                # this_span = point.fab_span_general()
                # если не создан участок переходим к прямому вызову
                yield from point.register_stage(Stage.INIT)

                if debug_is_on:
                    parent_span_context = schema_engine.adapter.keep_context.get('span')
                    logging.debug(schema_engine.schema_feed.variant)
                    logging.debug(f'Span context from task context: {parent_span_context}')

                if point.context.span:
                    yield from point.register_stage(Stage.PRE)

            except Exception as exc:
                # если здесь вылетели, то рабочий участок создать не удалось
                if point:
                    stop = point.register_own_error(exc)
                    if stop:
                        point = None
                log_tracer_error(exc, debug_is_on)

            finally:
                if point:
                    # в любом случае должны вызвать обернутую точку
                    if part == Part.OPEN and point and point.context and point.context.span:
                        open_span_id = point.context.span.span_instance.context.span_id
                        point.context.adapter.open_span_points[open_span_id] = point
                    try:
                        result = yield from point.run()
                    except (Exception, PostgreSQLError) as exc:
                        if point:
                            point.register_error(exc, sys.exc_info())
                        # фиксируем данные об исключении
                        if debug_is_on:
                            logging.exception(exc)
                        if reraise_exceptions:
                            raise exc

                    finally:
                        # так или иначе пробуем завершить формирование
                        # участка трассера
                        if part is None:
                            try:
                                if point and isinstance(result, TracerCollectedException):
                                    point.register_error(result.exc, (type(result.exc), result.exc,
                                                                      result.exc.__traceback__))
                                    result = result.result
                                yield from point.register_stage(Stage.POST)
                            except Exception as exc:
                                if point:
                                    point.register_error(exc, sys.exc_info())
                                    point.register_own_error(exc)
                                log_tracer_error(exc, debug_is_on)

                if not point or point.has_not_been_done():
                    # если вызов через аксессор не удался, вызываем напрямую
                    result = yield from point_coro(*args, **kwargs)

            return result

        return decorated

    return wrapper



class Component(object):
    def __init__(self):
        super(Component, self).__init__()
        self.loop = None
        self.app = None
        self.db = None
        self.config = None

    async def prepare(self):
        raise NotImplementedError()

    async def start(self):
        raise NotImplementedError()

    async def stop(self):
        raise NotImplementedError()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# noinspection PyProtectedMember

"""Проверка создания экземпляров обработчиков схем
"""

from tracuni.trac_uni import (
    SpanSide,
    APIKind,
    Stage,
)

import pytest


# pytest.skip('skipping {} tests'.format(__name__), allow_module_level=True)


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
def test_engine_instance(
    engine_types,
    simple_engine,
):
    """Создание процессора схемы

        * Создаётся экземпляр процессора
        * Инициализируется выбранным вариантом
        * Получает стандартную схему
        * Раскладывает схему для данного варианта
    """
    concrete_type = engine_types[0]
    p = simple_engine
    assert (
        isinstance(p, concrete_type)
    ), "Создаётся экземпляр процессора"
    assert (
        p.secret_mask_method_name == 'test_pipe_secret'
    ), "Есть корректный атрибут наименования модификатора маскирования"
    assert (
        p.skip_secret_destinations == []
    ), "Есть корректный атрибут списка отмены маскирования"
    assert (
        {p.method_to_stage.values()}
        ^
        {Stage.INIT, Stage.PRE, Stage.POST}
    ), "Есть корректный атрибут отображения наименований методов точки на фазы"
    assert (
        p.extractors
    ), "Есть непустой атрибут списков правил по фазам"


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
def test_feed_instance_no_custom_rules(
    engine_types,
    variants,
    mock_standard_schema,
):
    """Создание процессора схемы

        * Создаётся экземпляр процессора
        * Инициализируется выбранным вариантом
        * Получает стандартную схему
        * Раскладывает схему для данного варианта
    """
    feed_t = engine_types[2]
    c = feed_t(
        variants[0],
    )
    assert (
        isinstance(c, feed_t)
    ), "Создаётся экземпляр поставщика схем"
    assert (
        hasattr(c, 'variant')
    ), "В созданном экземпляре процессоре есть вариант"
    assert (
        c.variant.api_kind == APIKind.HTTP
    ), "В созданном экземпляре процессоре есть вариант переданного " \
       "вида интерфейса"
    assert (
        c.variant.span_side == SpanSide.IN
    ), "В созданном экземпляре процессоре есть вариант переданного " \
       "направления"
    assert (
        [
            (
                rule.destination,
                rule.stage,
                [o.section for o in rule.origins],
                len(rule.pipeline),
            )
            for ruleset in mock_standard_schema.values()
            for rule in ruleset
        ]
        ==
        [
            (
                rule.destination,
                rule.stage,
                [o.section for o in rule.origins],
                len(rule.pipeline),
            )
            for ruleset in c.standard_schema.values()
            for rule in ruleset
        ]
    ), "Импортируется нужная стандартная схема"


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
def test_feed_instance_custom_rules(
    engine_types,
    mock_standard_schema,
    both_custom_rules_engine,
):
    feed_t = engine_types[2]
    c = feed_t(
        both_custom_rules_engine.schema_feed.variant,
        both_custom_rules_engine.schema_feed.point_ruleset,
        both_custom_rules_engine.schema_feed.custom_schema,
    )
    assert (
        c.point_ruleset
        ==
        mock_standard_schema[tuple(mock_standard_schema.keys())[0]]
    ), "В объекте сохраняется пользовательская схема точки"
    assert (
        c.custom_schema
        ==
        mock_standard_schema
    ), "В объекте сохраняется пользовательская карта схем"


# @pytest.mark.skip(reason="get out of the other tests' way to debug them")
def test_engine_feed_no_variant(engine_types):
    e_t, _, _, err = engine_types
    expected_err = err.ProcessorInitNoneOrWrongTypeVariant
    with pytest.raises(expected_err):
        v = e_t(1)
        del v
    with pytest.raises(expected_err):
        # noinspection PyArgumentList
        v = e_t()  # noqa
        del v
    with pytest.raises(expected_err):
        v = e_t(variant=1)
        del v


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
def test_processor_wrong_schema(engine_types, variants):
    e_t, _, _, err = engine_types
    with pytest.raises(err.ProcessorInitWrongTypeCustomSchema):
        v = e_t(variant=variants[0], point_ruleset=1)
        del v
    with pytest.raises(err.ProcessorInitWrongTypeCustomSchema):
        v = e_t(variant=variants[0], custom_schema=1)
        del v



import aiozipkin.transport as azt
import logging
from urllib.parse import quote

from tracuni.side_steps.stat.adapter import StatsD
from tracuni.misc.select_coroutine import get_coroutine_decorator
from tracuni.define.const import STATSD_DURATION_FACTOR
async_decorator = get_coroutine_decorator()


class TracerTransport(azt.Transport):
    # noinspection PyProtectedMember
    def __init__(
            self,
            logging_conf,
            statsd_conf,
            loop,
            use_tornado=False,
    ):
        tracer_url = logging_conf.url
        send_interval = logging_conf.send_interval
        super(TracerTransport, self).__init__(
            tracer_url,
            send_interval=send_interval,
            loop=loop
        )
        self.loop = loop
        self.__tracer = logging_conf.name
        self.use_tornado = use_tornado

        statsd_conf = StatsD.read_statsd_configuration(statsd_conf)
        self.stats = None
        if statsd_conf.enable:
            self.stats = StatsD(statsd_conf._asdict())
            self.stats.connect()
        else:
            logging.warning("StatsD send has NOT been started")
        self.statsd_is_configured = statsd_conf.enable and self.stats
        if self.statsd_is_configured:
            logging.info("StatsD send is nominal")

    @async_decorator
    def _send(self):
        data = self._queue[:]

        try:
            if self.statsd_is_configured:
                yield self._send_to_statsd(data)
        except Exception as e:
            logging.exception(e)

        try:
            if self.__tracer == 'zipkin':
                if self.use_tornado:
                    yield super(TracerTransport, self)._send()
                else:
                    yield from super(TracerTransport, self)._send()
            else:
                self._queue = []
        except Exception as e:
            logging.exception(e)

    @async_decorator
    def _send_to_statsd(self, data):
        if self.stats:
            for rec in data:
                name = rec['name']
                tags = rec.get('tags', {})

                api_kind = name.split('::')[0]
                span_side = rec['kind']
                json_method = tags.get('data.method')
                json_response_code = str(tags.get('data.code'))
                widget_id = tags.get('app.widget')
                app_method = tags.get('app.method', '').split(':')[-1]
                rq_hostname = tags.get('rq.url.hostname')
                try:
                    rsp_status = int(tags.get('rsp.status') or 0)
                except ValueError:
                    rsp_status = 0
                duration = int(round(rec["duration"] / STATSD_DURATION_FACTOR))

                span_side_map = {
                    'SERVER': 'in',
                    'CLIENT': 'out',
                }

                if tags.get('error') is not None:
                    own = tags.get('error.own')
                    xls = tags.get('error.class')

                    if own or api_kind == '<unknown>':
                        variant = "tracer_own"
                    else:
                        variant = f"{api_kind}_{span_side_map.get(span_side)}"

                    # ALL:ALL if error
                    # counter <unit_name>.exception,class=x,variant=x
                    # class = exception class name
                    # variant = ((HTTP|AMQP|DB_IN|OUT)|tracer_own)
                    self.stats.exception_count_glork(xls, variant)

                if api_kind == 'http':
                    if span_side == 'CLIENT':
                        # HTTP:OUT
                        # timer <unit_name>.http_requests,code=x,hostname=x
                        # code = peer http response code
                        # hostname = peer address
                        self.stats.http_out_time_glork(
                            rsp_status, rq_hostname,
                            duration
                        )
                    elif span_side == 'SERVER':
                        # HTTP:IN
                        # counter <unit_name>.gen_response_code,code=x,widget_id=x,method=x
                        # code = API response code, where 0 = success
                        # widget_id = widget data from request
                        # method = API method from request
                        if json_response_code is not None:
                            self.stats.http_in_code_count_glork(
                                json_response_code, widget_id, json_method
                            )
                        # HTTP:IN
                        # timer <unit_name>.handler_timers,name=x,widget_id=x,method=x
                        # name = handler method name
                        # widget_id = widget data from request
                        # method = API method from request
                        self.stats.http_in_method_time_glork(
                            app_method, widget_id, json_method,
                            duration,
                        )

                if api_kind == 'db':
                    # DB:OUT
                    # timer <unit_name>.db_query_function,func_name=<method_name>
                    # func_name = db api method name
                    self.stats.db_time_glork(
                        app_method,
                        duration,
                    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Модуль доступа к вызываемой точке

"""
import logging
import traceback
import re
from collections import namedtuple
from copy import copy
from functools import partial
from inspect import signature, Parameter
from typing import (
    NamedTuple,
    Callable,
    Sequence,
    Dict,
    Any,
    MutableSequence,
)

import aiohttp

from tracuni.point.span_general import SpanGeneral
from tracuni.misc.helper import compose_own_call_stack
from tracuni.define.const import (
    PATH_IGNORE_METHODS,
    PATH_IGNORE_PIECE,
    COLLECT_TRACE,
)
from tracuni.define.type import (
    PointContext,
    Stage,
    Destination,
    Variant,
    SpanSide,
    APIKind,
)
import tracuni.define.errors as err
from tracuni.schema.pipe_methods import pipe_mask_secret_catch_essentials
from tracuni.misc.select_coroutine import get_coroutine_decorator, ASYNC_VARIANT
async_decorator = get_coroutine_decorator()


class PointAccessor:
    """Доступ к точке вызова и связанным с ней данным

    * Хранит контекст: ссылки на точку, аргументы, сокращённый стек вызовов
    и схему варианта точки
    * Преобразует все аргументы к именованным, основываясь на сигнатуре
    точки и переданным именованным аргументам для упрощения доступа к ним
    * Копирует записываемое в трассер содержимое аргументов перед их
    модификацией
    * Регистрирует сигнал о достижении определённой фазы, вызывает извлечение
    данных этой фазы по схеме варианта
    * Хранит извлеченные данные в определённой структуре, предназначенной
    для записи в трассер и для переиспользования на последующих шагах работы
    экстрактора
    * Модифицирует ограниченное множество аргументов (заголовки HTTP)
    * Вызывает точку с оригинальными и модифицированными аргументами
    """

    def __init__(self,
                 original_args: Sequence,
                 original_kwargs: Dict,
                 point_context: PointContext,
                 pass_itself_as_kwargs_item=False,
        ):
        self._positional_names = []
        self._original_args = original_args
        self._original_kwargs = original_kwargs
        self._applied_point = None
        self._done = False
        self.pass_itself_as_kwargs_item = pass_itself_as_kwargs_item

        point_args = self._compose_kwargs(point_context.point_ref)

        # noinspection PyProtectedMember
        self.context = PointContext(**{
            **point_context._asdict(),
            "point_args": point_args,
        })
        self._span_general = SpanGeneral(
            self.context.engine.schema_feed.variant.span_side,
            self,
        )
        if not self._span_general:
            raise err.NoSpanException

    @classmethod
    def fab_default(cls, args, kwargs, point_coro, schema_engine, remote_name, pass_itself_as_kwargs_item=False):
        return cls(
            args,
            kwargs,
            PointContext(
                point_ref=point_coro,
                engine=schema_engine,
                adapter=schema_engine.adapter,
                client=schema_engine.adapter.client,
                reuse={
                    "custom_remote_endpoint_name": remote_name,
                    "error": None,
                    "should_not_trace": False,
                },
                debug={},
                headers={},
                span_name={},
                span_tags={},
                span_logs=[],
            ),
            pass_itself_as_kwargs_item=pass_itself_as_kwargs_item,
        )

    def stack_frames_to_context(self, current_frame):
        point_coro = self.context.point_ref
        stack_frames = (
            # первым элементом - название вызываемого метода
            {
                'full_path': point_coro.__qualname__,
                'short_path': point_coro.__name__,
            },
        )
        if COLLECT_TRACE:
            # собрать данные по стэку вызовов в данной точке
            # inspect.getouterframes(inspect.currentframe())

            frame_list = []
            frame = current_frame
            while frame:
                frame_list.append(frame)
                frame = frame.f_back
            stack_frames = (
                # добавить название вызываемого метода
                *stack_frames,
                *compose_own_call_stack(
                    PATH_IGNORE_METHODS,  # пропустить методы с именами
                    PATH_IGNORE_PIECE,  # пропустить модули по такому пути
                    result_length=5,  # общий размер списка не превышает
                    from_index=1,  # начинать просмотр стэка с позиции
                    look_up_stop_idx=10,  # не просматривать дальше позиции
                    frame_list=frame_list,  # стэк
                )
            )
        # noinspection PyProtectedMember
        self.context = PointContext(**{
            **self.context._asdict(),
            "call_stack": stack_frames,
        })

    def apply_point(self):
        kwargs = copy(self._original_kwargs)
        # noinspection PyProtectedMember
        kwargs.update(self.context.point_args._asdict())
        kwargs.pop('method_arguments_without_names')
        args = list(self._original_args)
        for idx, pos_name in enumerate(
            self._positional_names[:len(self._original_args)]
        ):
            args[idx] = kwargs.pop(pos_name, None)

        if self.pass_itself_as_kwargs_item:
            kwargs['tracuni_point'] = self

        self._applied_point = partial(self.context.point_ref, *args, **kwargs)

    @async_decorator
    def run(self):
        result = None
        if self.context.span and self._applied_point:
            try:
                variant = self.context.engine.schema_feed.variant
                if (
                    ASYNC_VARIANT == "TORNADO"
                    and
                    variant == Variant(SpanSide.OUT, APIKind.INNER_RETRY)
                ):
                    span_context = self.context.adapter.keep_context
                    result = yield from span_context.run_root_with_context(coro=self._applied_point)
                else:
                    result = yield from self._applied_point()
            finally:
                self._done = True
            self.update_context({"point_result": result})
        return result

    def has_not_been_done(self):
        return not self._done

    @async_decorator
    def register_stage(self, stage: Stage):
        yield from self._span_general.run_stage(stage)

    @async_decorator
    def register_error(self, exc, exc_info):
        found_exc_name = re.search(r"<class '(.*)'>", str(type(exc))).groups()
        if found_exc_name:
            found_exc_name = found_exc_name[0].split('.')[-1]
        else:
            found_exc_name = ''
        error = {
            "error_class": found_exc_name,
            "error_message": pipe_mask_secret_catch_essentials(str(exc))[0],
            "error_info": pipe_mask_secret_catch_essentials(
                '\n'.join(traceback.format_exception(*exc_info)),
            )[0],
        }
        self._span_general.register_error(error, getattr(exc, 'error_code', None))
        return error

    def register_own_error(self, exc):
        if isinstance(
            exc,
            (err.HTTPPathSkip, err.SpanNoParentException)
        ):
            return True
        else:
            self._span_general.emergency_exit(exc)

    def update_context(self, new_data):
        # noinspection PyProtectedMember
        self.context = PointContext(**{
            **self.context._asdict(),
            **new_data,
        })

    def update_context_by_destination(
        self,
        destination: Destination,
        val: Any,
    ):
        """

        Parameters
        ----------
        destination
            Данные по назначению записи
        val
            Значение которое надо отправить в контекст
        """

        # noinspection PyProtectedMember
        dest = self.context._asdict().get(
            destination.section.name.lower()
        )
        if dest is None:
            return

        if destination.section.name.upper() == 'POINT_ARGS':
            if destination.name == 'request':
                if isinstance(getattr(dest, "request", None), aiohttp.web.Request):
                    setattr(dest.request, "zipkin_headers", val)
                    return


        if isinstance(dest, MutableSequence):
            dest.append(val)

        elif isinstance(dest, dict):
            if isinstance(destination.name, dict):
                dest.update(val)
            else:
                dest[destination.name] = val

        elif "PointArguments" in str(type(dest)):
            # noinspection PyProtectedMember
            old_vals = dest._asdict()
            new_args = namedtuple("PointArguments", old_vals)(**{
                **old_vals,
                destination.name: {
                    **(old_vals.get(
                        destination.name,
                        {}
                    ) or {}),
                    **val,
                }
            })
            self.update_context({"point_args": new_args})

        else:
            try:
                if destination.name:
                    setattr(dest, destination.name, val)
            except AttributeError:
                raise err.ProcessorExtractCantSetDestinationAttr

    def _compose_kwargs(self, point_ref: Callable) -> NamedTuple:
        sig = signature(point_ref).parameters
        self._positional_names = arg_names = [
            el
            for el in sig
            if el not in ('args', 'kwargs')
        ]
        kwargs = copy(self._original_kwargs)
        args = copy(self._original_args)
        kwargs.update(zip(arg_names, args))

        param_desc = sig
        for param in param_desc:
            pass_value = kwargs.get(param, param_desc.get(param).default)
            if pass_value != Parameter.empty:
                kwargs[param] = pass_value

        kwargs_l = len(kwargs)
        kwargs.update({'method_arguments_without_names': None})
        if kwargs_l < len(args):
            kwargs['method_arguments_without_names'] = args[kwargs_l:]

        point_args = namedtuple(
            'PointArguments',
            kwargs
        )(**kwargs)

        return point_args



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Тестовые данные для внутренних структур обработчика схем
"""

import pytest

from tracuni.trac_uni import (
    Variant,
    SpanSide,
    APIKind,
    Rule,
    Destination,
    DestinationSection,
    Origin,
    OriginSection,
    StageToSections,
    Stage,
    PointContext,
)
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from tracuni.trac_uni import RuleSet  # noqa


@pytest.fixture
def variants():
    variant_sample_one = Variant(
        span_side=SpanSide.IN,
        api_kind=APIKind.HTTP,
    )
    variant_sample_two = Variant(
        span_side=SpanSide.OUT,
        api_kind=APIKind.HTTP,
    )
    variant_sample_all_all = Variant(
        span_side=SpanSide.ALL,
        api_kind=APIKind.ALL,
    )
    variant_sample_all_api_kinds = Variant(
        span_side=SpanSide.IN,
        api_kind=APIKind.ALL,
    )
    variant_sample_all_span_sides = Variant(
        span_side=SpanSide.ALL,
        api_kind=APIKind.HTTP,
    )
    return (
        variant_sample_one,
        variant_sample_two,
        variant_sample_all_all,
        variant_sample_all_api_kinds,
        variant_sample_all_span_sides,
    )


@pytest.fixture
def stage_to_sections() -> StageToSections:
    return {
        Stage.INIT: [
            DestinationSection.REUSE,
        ],
        Stage.POST: [
            OriginSection.POINT_RESULT,
        ],
    }


@pytest.fixture
def point_ruleset():
    ruleset = (
        Rule(
            description="Уцелеет только один",
            destination=Destination(
                section=DestinationSection.LOGS,
                name='None'
            ),
            origins=(Origin(
                section=OriginSection.CLIENT, getter=lambda x: x
            ),),
            pipeline=(),
        ),
        Rule(
            destination=Destination(
                section=DestinationSection.LOGS,
                name='None'
            ),
            origins=(Origin(
                section=OriginSection.CLIENT, getter=lambda x: x
            ),),
            pipeline=(),
        ),
        Rule(
            destination=Destination(
                section=DestinationSection.LOGS,
                name='None'
            ),
            origins=(Origin(
                section=OriginSection.CLIENT, getter=lambda x: x
            ),),
            pipeline=(),
        ),
        Rule(
            destination=Destination(
                section=DestinationSection.TAGS,
                name='None'
            ),
            origins=(Origin(
                section=OriginSection.CLIENT, getter=lambda x: x
            ),),
            pipeline=(),
        ),
    )  # type: RuleSet
    invalid_ruleset = ('wrong',) + ruleset
    return (
        ruleset,
        invalid_ruleset,
    )


@pytest.fixture
def gen_schema_rule():
    def fn(
        description="",
        destination=None,
        origins=None,
        pipeline=None,
        stage=Stage.INIT,
    ):
        if destination is None:
            destination = Destination(
                section=DestinationSection.LOGS,
                name='None'
            )
        if origins is None:
            origins = (
                Origin(section=OriginSection.CLIENT, getter=lambda x: x),
            )
        if pipeline is None:
            pipeline = (lambda x: x,)
        return Rule(**locals())
    return fn


@pytest.fixture
def select_variants_schema_map(gen_schema_rule, variants):
    v1, v2, v_all_all, v_all_api_kinds, v_all_span_sides, *_ = variants

    return {
        v1: (gen_schema_rule('1a'), gen_schema_rule('1b'),),
        v_all_api_kinds: (gen_schema_rule('2'),),
        v_all_span_sides: (gen_schema_rule('3a'), gen_schema_rule('3b'),),
        v_all_all: (gen_schema_rule('4'),),
        v2: (gen_schema_rule('5a'), gen_schema_rule('5b'),),
    }


@pytest.fixture
def point_context():
    class Data:
        some = "value"

    def fn(x):
        return x

    res = PointContext(
        point_args={"some": "value"},
        point_ref=fn,
        adapter=object(),
        engine=object(),
        reuse={"some": "value"},
        span=Data(),
        logs=[],
        tags={},
        span_id={},
    )
    return res



import asyncio
import signal
import logging

from aiohttp.web_runner import _raise_graceful_exit

from app.core.component import Component
from app.core.helper import GracefulExit, PrepareError


class BaseApp(object):
    def __init__(self, config=None, loop=None):
        super(BaseApp, self).__init__()
        self.config = config
        self.loop = loop or asyncio.get_event_loop()
        self._components: dict = {}
        self._stop_deps: dict = {}
        self._stopped: list = []
        self._tracer: az.Tracer = None

    def attach_component(
            self,
            name: str,
            comp: Component,
            stop_after: list = None
    ):
        if not isinstance(comp, Component):
            raise UserWarning()
        if name in self._components:
            raise UserWarning()
        comp.loop = self.loop
        comp.app = self
        self._components[name] = comp
        self._stop_deps[name] = stop_after

    def __getattr__(self, item):
        if item not in self._components:
            raise AttributeError
        return self._components[item]

    @staticmethod
    def log_err(err):
        if isinstance(err, BaseException):
            logging.exception(err)
        else:
            logging.error(err)

    @staticmethod
    def log_warn(warn):
        logging.warning(warn)

    @staticmethod
    def log_info(info):
        logging.info(info)

    @staticmethod
    def log_debug(debug):
        logging.debug(debug)

    def setup_logging(
            self,
            tracer_driver,
            tracer_svc_name,
            tracer_url,
            statsd_addr,
            statsd_prefix,
            sample_rate=1.0,
            send_interval=3
    ):
        self._tracer = None

    async def _shutdown_tracer(self):
        if self._tracer:
            self.log_info("Shutting down tracer")
            await self._tracer.close()

    @property
    def tracer(self):
        return self._tracer

    def run(self):
        try:
            self.loop.run_until_complete(self.run_prepare())
        except PrepareError as e:
            self.log_err(e)
            return 1
        except KeyboardInterrupt:  # pragma: no cover
            return 1
        self.run_loop()
        self.loop.run_until_complete(self.run_shutdown())
        print("Bye")
        if hasattr(self.loop, 'shutdown_asyncgens'):
            self.loop.run_until_complete(self.loop.shutdown_asyncgens())
        self.loop.close()
        return 0

    async def run_prepare(self):
        self.log_info('Prepare for start')
        await asyncio.gather(*[comp.prepare() for comp in self._components.values()], loop=self.loop)

        self.log_info('Starting...')
        await asyncio.gather(*[comp.start() for comp in self._components.values()], loop=self.loop)
        self.log_info('Running...')

    def run_loop(self):
        try:
            self.loop.add_signal_handler(signal.SIGINT, _raise_graceful_exit)
            self.loop.add_signal_handler(signal.SIGTERM, _raise_graceful_exit)
        except NotImplementedError:  # pragma: no cover
            # add_signal_handler is not implemented on Windows
            pass
        try:
            self.loop.run_forever()
        except GracefulExit:  # pragma: no cover
            pass

    async def run_shutdown(self):
        self.log_info('Shutting down...')
        for comp_name in self._components:
            await self._stop_comp(comp_name)
        await self._shutdown_tracer()

    async def _stop_comp(self, name):
        if name in self._stopped:
            return
        if name in self._stop_deps and self._stop_deps[name]:
            for dep_name in self._stop_deps[name]:
                await self._stop_comp(dep_name)
        await self._components[name].stop()
        self._stopped.append(name)



from functools import partial

from tracuni.define.type import (
    Rule,
    Destination,
    DestinationSection,
    Origin,
    OriginSection,
    Stage,
    PipeCommand,
)
from tracuni.schema.pipe_methods import (
    ext_tracer_out_point,
    ext_tracer_path,
    ext_tracer_full_path,
    pipe_head,
    pipe_sep_string,
    pipe_mask_secret_catch_essentials,
    pipe_dump,
    pipe_cut,
    ext_api_kind,
)
from tracuni.schema.shared_rules import rule_amqp_out_tracer_context

ruleset = (
    rule_amqp_out_tracer_context,
    Rule(
        description="Сохранить тело сообщения",
        destination=Destination(DestinationSection.REUSE, 'message'),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(
                OriginSection.POINT_ARGS, "message",
            ),
        ),
        stage=Stage.INIT,
    ),
    Rule(
        description="Путь по методам, вызвавшим запрос",
        destination=Destination(DestinationSection.SPAN_TAGS, 'app.path'),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(OriginSection.CALL_STACK, ext_tracer_path),
        ),
        stage=Stage.INIT,
    ),
    Rule(
        description="Метод, вызвавший запрос",
        destination=Destination(DestinationSection.SPAN_TAGS, 'app.method'),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(OriginSection.CALL_STACK, ext_tracer_full_path),
        ),
        stage=Stage.INIT,
    ),
    Rule(
        description="Задать имя участку",
        destination=Destination(
            DestinationSection.SPAN_NAME,
            'name',
          ),
        pipeline=(
            lambda x: ("amqp", x[0]['method'].lower() or x[1], x[0]['url_out'].lower()),
            partial(pipe_sep_string, **{'sep': ('::', ' (', ')')}),
        ),
        origins=(
            Origin(OriginSection.POINT_ARGS, 'message'),
            Origin(OriginSection.POINT_ARGS, 'routing_key'),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Наименование партнерского сервиса сохранить для "
                    "переиспользования",
        destination=Destination(
            DestinationSection.REUSE,
            'peer_name',
          ),
        pipeline=(
            pipe_head,
            lambda v: v.get('exchange', v.get('exchange_name')),
        ),
        origins=(
            Origin(OriginSection.POINT_ARGS, 'config'),
        ),
        stage=Stage.INIT,
    ),
    Rule(
        description="Наименование партнерского сервиса записать в участок",
        destination=Destination(
            DestinationSection.SPAN_NAME,
            'remote_endpoint',
          ),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(OriginSection.REUSE, 'peer_name'),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Вывести тело сообщения в аннотации",
        destination=Destination(DestinationSection.SPAN_LOGS, 'message'),
        pipeline=(
            pipe_head,
            pipe_dump,
            pipe_cut,
            pipe_mask_secret_catch_essentials,
            lambda x: x[1],
            PipeCommand.TEE,
            Destination(DestinationSection.SPAN_TAGS, {}),
            pipe_head,
        ),
        origins=(
            Origin(
                OriginSection.REUSE, "message",
            ),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Запомнить ошибку для вывода",
        destination=Destination(DestinationSection.REUSE, 'error'),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(OriginSection.SPAN, 'error'),
        ),
        stage=Stage.POST,
    ),
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# noinspection PyProtectedMember

"""Проверка поставщика схем
"""

from collections import OrderedDict

from tracuni.trac_uni import (
    Destination,
    DestinationSection,
    PipeCommand,
    Stage,
)

# import pytest

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from tracuni.trac_uni import Extractors  # noqa


# pytest.skip('skipping {} tests'.format(__name__), allow_module_level=True)


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
def test__compose_variants(
    simple_engine,
    custom_plus_standard_schema_map,
):
    p = simple_engine

    custom_plus_standard_composed = tuple(
        p.schema_feed._compose_variants(el, tuple()) for el in
        custom_plus_standard_schema_map
    )

    schema_names = ''.join(el.description for el in (
        *custom_plus_standard_composed[0],
        *custom_plus_standard_composed[1],
    ))

    assert (
        schema_names == 'c1ac1bc2c3ac3bc4s1as1bs2s3as3bs4'
    ), "Работает отбор правил схем по варианту"


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
def test__group_by_stage(
    simple_engine,
    auto_stage_rules,
    stage_to_sections
):
    p = simple_engine

    assert (
        p.schema_feed.stage_to_sections == stage_to_sections
    ), "Используются нужные подставные данные по отношению секций к фазам"
    schema = tuple(
        el
        for tuple_el
        in tuple(OrderedDict(sorted(
            auto_stage_rules.items(), key=lambda x: x[0]
        )).values())
        for el in tuple_el
    )
    stage_to_schema = p.schema_feed._distribute_by_stages(schema)
    assert (
        len(stage_to_schema) == 3
    ), "Все фазы в отношении на месте"
    assert (
        not set(''.join(
            el.destination.name
            for el in stage_to_schema.get(Stage.INIT)
        )) ^ set('ab')
    ), "Правила фазы старта попали на своё место без лишних"
    assert (
        not set(''.join(
            el.destination.name
            for el in stage_to_schema.get(Stage.PRE)
        )) ^ set('cd')
    ), "Правила фазы вызова попали на своё место без лишних"
    assert (
        not set(''.join(
            el.destination.name
            for el in stage_to_schema.get(Stage.POST)
        )) ^ set('efg')
    ), "Правила фазы завершения попали на своё место без лишних"


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
def test__should_skip_level_below(
    simple_engine,
    gen_schema_rule,
):
    flag = simple_engine.schema_feed._should_skip_level_below()[1]
    assert bool(flag) is False,\
        "Пустая схема не содержит флага"
    flag = simple_engine.schema_feed._should_skip_level_below((
        gen_schema_rule('should_skip', Destination(
                section=DestinationSection.ALL
            ),
            pipeline=(PipeCommand.SKIP_LEVELS_BELOW,)
        ),
    ))[1]
    assert bool(flag) is True,\
        "Схема содержит флаг"
    flag = simple_engine.schema_feed._should_skip_level_below((
        gen_schema_rule('no_skip', Destination(
            section=DestinationSection.LOGS,
            name='no_skip'
        )),
    ))[1]
    assert bool(flag) is False,\
        "Схема не содержит флага"


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
def test_skip_all_levels_below(
    engine_types,
    gen_schema_rule,
    custom_plus_standard_schema_map,
    variants,
):
    p_t, *_ = engine_types
    p = p_t(
        variants[0],
        (
            gen_schema_rule('pc1'),
            gen_schema_rule('pc2', Destination(
                section=DestinationSection.POINT_ARGS,
                name='ARGS'
            )),
            gen_schema_rule('pc3'),
            gen_schema_rule('pc4skip', Destination(
                    section=DestinationSection.ALL
                ),
                pipeline=(PipeCommand.SKIP_LEVELS_BELOW,)
            ),
        ),
        custom_plus_standard_schema_map[0],
    )
    assert (
        not
        {
            el.description
            for tuple_el in p.extractors.values()
            for el in tuple_el
        }
        ^
        {'pc1', 'pc2'}
    ), "Если указана команда пропуска нижестоящих уровней в пользовательской" \
       "схеме точки, остаются только правила из пользовательской схемы"
    assert (not len(tuple(
        pipe
        for _, schema in p.extractors.items()
        for extractor in schema
        for pipe in extractor.pipeline
        if pipe == PipeCommand.SKIP_LEVELS_BELOW
    ))), "Правило-команда по игнорированию менее приоритетных должно быть " \
         "убрано из схемы после его применения"
    parsed_twice = p.schema_feed.parse()  # type: Extractors
    assert (
        not
        {
            el.description
            for tuple_el in parsed_twice.values()
            for el in tuple_el
        }
        ^
        {'pc1', 'pc2'}
    ), "И это, минутку внимания, должно быть идемпотентно"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# noinspection PyProtectedMember

"""Тестовые данные обработчика в виде полных схем
"""
from itertools import groupby
from operator import itemgetter

import pytest

from tracuni.define.type import (
    Rule,
    Destination,
    DestinationSection,
    Origin,
    OriginSection,
    Stage,
)


@pytest.fixture
def custom_plus_standard_schema_map(select_variants_schema_map):
    """Тестовые схемы для проверки композиции

        Генерирует две схемы одну имитирующую пользовательскую, другую
        стандартную, отличающиеся только описанием элементов: к ним
        добавляется префикс "c" или "s" соответственно. Все элементы имеют
        одинаковое назначение (<section>.NONE, 'None'), соответственно при
        композиции должен браться только первый из них. Исключение, элементы
        с индексом единица в каждой схеме (то есть вторые). В этом случае
        для пользовательской назначение меняется на (<section>.TAGS,
        'tags'), а для стандартной на (<section>.LOGS, 'logs'). Это приводит
        к тому, что первые такие должны оставаться в итоговой схеме,
        как обладающие уникальными значениями назначения.
    """
    schema_levels = ('c', 's')
    different_destinations = {
        schema_levels[0]: Destination(DestinationSection.SPAN_TAGS, 'tags'),
        schema_levels[1]: Destination(DestinationSection.SPAN_LOGS, 'logs'),
    }

    def schema_dict(schema_item):
        # noinspection PyProtectedMember
        return schema_item._asdict()

    return tuple(dict(
        (k, tuple(el[1] for el in v))
        for k, v in groupby(
            (
                (
                    variant,
                    Rule(**{
                        **schema_dict(schema_item),
                        **{
                            'description': level + schema_item.description,
                            'destination': (
                                schema_item.destination
                                if idx != 1
                                else different_destinations[level]
                            )
                        }
                    })
                )
                for variant, schema in select_variants_schema_map.items()
                for idx, schema_item in enumerate(schema)
            ),
            key=itemgetter(0)
        )
    ) for level in schema_levels)


@pytest.fixture
def mock_standard_schema(variants, point_ruleset, gen_schema_rule):
    return {
        variants[0]: point_ruleset[0],
        variants[1]: (
            gen_schema_rule(
                destination=Destination(
                    DestinationSection.REUSE,
                    'test_override',
                ),
                origins=(Origin(
                    OriginSection.POINT_ARGS,
                    lambda _: "standard",
                ),),
                pipeline=(lambda x: x[0],),
            ),
            gen_schema_rule(
                destination=Destination(
                    DestinationSection.SPAN_LOGS,
                    'test_override',
                ),
                origins=(Origin(
                    OriginSection.POINT_ARGS,
                    lambda _: "standard",
                ),),
                pipeline=(lambda x: x[0],),
            ),
        )
    }


@pytest.fixture
def auto_stage_rules(gen_schema_rule):
    return {
        Stage.INIT: (
            gen_schema_rule(
                'явно указана фаза',
                stage=Stage.INIT,
                destination=Destination(
                    section=DestinationSection.SPAN_LOGS,
                    name='a',
                ),
            ),
            gen_schema_rule(
                'нет относящих к POST, есть к INIT',
                destination=Destination(
                    section=DestinationSection.REUSE,
                    name='b',
                ),
                stage=None,
            ),
        ),
        Stage.PRE: (
            gen_schema_rule(
                'явно указана фаза',
                stage=Stage.PRE,
                destination=Destination(
                    section=DestinationSection.SPAN_LOGS,
                    name='c',
                ),
            ),
            gen_schema_rule(
                'нет ни относящих к POST, ни к INIT',
                destination=Destination(
                    section=DestinationSection.SPAN_TAGS,
                    name='d',
                ),
                origins=(
                    Origin(
                        section=OriginSection.CLIENT,
                        getter=lambda x: x
                    ),
                ),
                stage=None,
            ),
        ),
        Stage.POST: (
            gen_schema_rule(
                'явно указана фаза',
                stage=Stage.POST,
                destination=Destination(
                    section=DestinationSection.SPAN_LOGS,
                    name='e',
                ),
            ),
            gen_schema_rule(
                'есть и к POST, и к INIT',
                destination=Destination(
                    section=DestinationSection.REUSE,
                    name='f',
                ),
                origins=(
                    Origin(
                        section=OriginSection.POINT_RESULT,
                        getter=lambda x: x
                    ),
                ),
                stage=None,
            ),
            gen_schema_rule(
                'есть только к POST',
                destination=Destination(
                    section=DestinationSection.REUSE,
                    name='g',
                ),
                origins=(
                    Origin(
                        section=OriginSection.POINT_RESULT,
                        getter=lambda x: x
                    ),
                ),
                stage=None,
            ),
            gen_schema_rule(
                'Дубль назначения для проверки отбрасывания',
                destination=Destination(
                    section=DestinationSection.SPAN_LOGS,
                    name='c',
                ),
                origins=(
                    Origin(
                        section=OriginSection.POINT_RESULT,
                        getter=lambda x: x
                    ),
                ),
                stage=None,
            ),
        ),
    }



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Узел подключения к другим компонентам
        * обеспечивает взаимодействие с главным компонентом клиентского
        приложения и инициализируется вмете с ним;
        * хранит ссылку на главный компонент клиентского приложения;
        * получает, обрабатывает и сохраняет настройки журналирования;
        * создаёт драйверы подключения к внешним ресурсам журналирования;
        * выбирает варианты реализации асинхронности (для работы с aiohttp
         и для работы с Tornado);
        * уведомляет остальные компоненты библиотеки о своей готовности;

        .side_steps.stat.intercept.TracerTransport
            параллельная запись метрик
        .misc.helper.async_decorator,
            для подключения к Tornado
                    .HTTPServer,
                    .IOLoop,
                    .tornado_available,
        .misc.keep_tornado.context_accessor,
            сохранение родительского участка, используя механизм Tornado
"""
import logging
from abc import abstractmethod
from functools import partial

tornado_available = False
context_accessor = None

from tracuni.misc.select_coroutine import (
    get_coroutine_decorator,
    switch_to_tornado,
    HTTPServer,
    IOLoop,
)
# use only if tornado lib is available
if HTTPServer:
    from .misc.keep_tornado import context_accessor
    tornado_available = bool(switch_to_tornado())

from tracuni.adapter import (
    fab_tracer,
    FutureTracer,
)
import tracuni.define.errors as err

async_decorator = get_coroutine_decorator()


def init_tornado_loop(cls, port, app_conf, start_loop=False):
    # if HTTPServer:
    #     global context_accessor
    #     from .misc.keep_tornado import context_accessor
    #     tornado_available = bool(switch_to_tornado())
    if not tornado_available:
        raise err.TornadoNotAvailable

    logging_conf = app_conf.get('logging_conf')
    IOLoop.configure('tornado.platform.asyncio.AsyncIOLoop')
    this_app = cls(**app_conf)

    logging.info('%s listens at: %s' % (cls.__name__, port))

    tracer = fab_tracer(
        tracer_base_type=TracerBase
    )(
        this_app,
        logging_conf,
        app_conf.get('statsd_conf'),
        app_conf.get('system_conf'),
    )

    if tracer.config.enable:
        @async_decorator
        def prehook(request):
            context_accessor.run_root_with_context(
                coro=partial(this_app, request)
            )

        HTTPServer(prehook).listen(port)
    else:
        HTTPServer(this_app).listen(port)

    tracer.run()
    if start_loop:
        IOLoop.current().start()

    return


class TracerBase(FutureTracer):
    variant = 'tornado'

    def __init__(self):
        self.client = None
        super().__init__()
        self.log_info = None
        self.log_debug = None
        self.loop = IOLoop.current().asyncio_loop
        self.keep_context = context_accessor

    def run(self):
        self.log_info = getattr(
            self.client, 'log_info', logging.info
        )
        self.log_debug = getattr(
            self.client, 'log_debug', logging.debug
        )
        self.prepare_tracer()
        self.start_tracer()

    @abstractmethod
    def prepare_tracer(self):
        pass

    @abstractmethod
    def start_tracer(self):
        pass



import asyncio
from functools import partial

import aiohttp_jinja2
import jinja2
from aiohttp import web, web_runner, helpers
import traceback

from app.core.handler import BaseHandler
from app.core.helper import annotate_bytes
from .app import Component
import logging

access_logger = logging.getLogger('aiohttp.access')
SPAN_KEY = 'zipkin_span'

from tracuni import (
    tracer_point,
    Variant,
    SpanSide,
    APIKind,
)


class HttpServer(Component):
    def __init__(
            self,
            app,
            host,
            port,
            handler,
            shutdown_timeout=60.0,
            skip_trace=None,
            jinja_templates=None
    ):
        if not issubclass(handler, BaseHandler):
            raise UserWarning()
        super(HttpServer, self).__init__()
        self.app = app
        self.loop = app.loop
        self.web_app = web.Application(loop=self.loop, middlewares=[self.wrap_middleware, ])
        if jinja_templates:
            aiohttp_jinja2.setup(self.web_app, loader=jinja2.FileSystemLoader(jinja_templates))
        self.host = host
        self.port = port
        self.error_handler = None
        self.web_app_handler = None
        self.shutdown_timeout = shutdown_timeout
        self.uris = None
        self.handler = handler(self)
        self._sites: list = []
        self._runner: web_runner.AppRunner = None
        self.skip_trace = skip_trace or []

    async def wrap_middleware(self, app, handler):
        async def middleware_handler(request):
            if self.app.tracer:
                if request.path not in self.skip_trace:
                    context = az.make_context(request.headers)
                    if context is None:
                        sampled = azah.parse_sampled(request.headers)
                        debug = azah.parse_debug(request.headers)
                        span = self.app.tracer.new_trace(sampled=sampled, debug=debug)
                    else:
                        span = self.app.tracer.join_span(context)
                    request[SPAN_KEY] = span

                    with span:
                        span_name = '{0} {1}'.format(request.method.upper(), request.path)
                        span.name(span_name)
                        span.kind(azah.SERVER)
                        span.tag(HTTP_PATH, request.path)
                        span.tag(HTTP_METHOD, request.method.upper())
                        annotate_bytes(span, await request.read())
                        resp, trace_str = await self._handle(span, request, handler)
                        if trace_str is not None:
                            annotate_bytes(span, trace_str)
                        else:
                            if isinstance(resp, web.Response):
                                span.tag(azah.HTTP_STATUS_CODE, resp.status)
                                annotate_bytes(span, resp.body)
                        return resp
                else:
                    resp, trace_str = await self._handle(None, request, handler)
                    return resp
            else:
                resp, trace_str = await self._handle(None, request, handler)
                return resp

        return middleware_handler

    @tracer_point(Variant(SpanSide.IN, APIKind.HTTP), endpoint_name="custom")
    async def _handle(self, span, request, handler):
        try:
            resp = await handler(request)
            return resp, None
        except Exception as herr:
            trace = traceback.format_exc()

            if span:
                span.tag('error', 'true')
                span.tag('error.message', str(herr))
                span.annotate(trace)

            if self.error_handler:
                try:
                    resp = await self.error_handler(span, request, herr)
                except Exception as eerr:
                    if isinstance(eerr, web.HTTPException):
                        resp = eerr
                    else:
                        self.app.log_err(eerr)
                        resp = web.Response(status=500, text='')
                    trace = traceback.format_exc()
                    if span:
                        span.annotate(trace)
            else:
                if isinstance(herr, web.HTTPException):
                    resp = herr
                else:
                    resp = web.Response(status=500, text='')

            return resp, trace

    def add_route(self, method, uri, handler):
        if not asyncio.iscoroutinefunction(handler):
            raise UserWarning('handler must be coroutine function')
        self.web_app.router.add_route(method, uri, partial(self._handle_request, handler))

    def add_static(self, prefix, uri):
        self.web_app.router.add_static(prefix, uri)

    def set_error_handler(self, handler):
        if not asyncio.iscoroutinefunction(handler):
            raise UserWarning('handler must be coroutine function')
        self.error_handler = handler

    @staticmethod
    async def _handle_request(handler, request):
        res = await handler(request.get(SPAN_KEY), request)
        return res

    async def prepare(self):
        self.app.log_info("Preparing to start http server")
        self._runner = web_runner.AppRunner(
            self.web_app,
            handle_signals=True
        )

        await self._runner.setup()

        self._sites = []
        self._sites.append(web_runner.TCPSite(
            self._runner,
            self.host,
            self.port,
            shutdown_timeout=self.shutdown_timeout,
            ssl_context=None,
            backlog=128))

    async def start(self):
        self.app.log_info("Starting http server")
        await asyncio.gather(*[site.start() for site in self._sites], loop=self.loop)
        self.app.log_info('HTTP server ready to handle connections on %s:%s' % (self.host, self.port))

    async def stop(self):
        self.app.log_info("Stopping http server")
        if self._runner:
            await self._runner.cleanup()



import re
from typing import (
    Sequence,
    Dict,
)

from tracuni.define.const import (
    TAG_ESSENTIAL_FIELD_NAMES,
    TAG_ESSENTIAL_BODY_NAMES,
)
from tracuni.misc.helper import to_iter


def _render_essential_patterns(tag_essential_field_names):
    """Создание шаблонов для поиска существенных значений

    Parameters
    ----------
    tag_essential_field_names

    Returns
    -------

    """
    ess_flat_names = [
        name
        for el in tag_essential_field_names
        for name in to_iter(el)
    ]
    ess_first_name_map = {}
    for single_field_names in tag_essential_field_names:
        if (
            isinstance(single_field_names, Sequence)
            and
            len(single_field_names) > 1
        ):
            for name in single_field_names:
                ess_first_name_map[name] = single_field_names[0]

    ess_patterns = {}
    regexp_string = r'[\'"]{}[\'"]\s*:\s*[\'"]?(.+?)\s*[\}},\'"]'
    for tag_field_name in ess_flat_names:
        ess_patterns[tag_field_name] = re.compile(
            regexp_string.format(tag_field_name)
        )

    return ess_patterns, ess_flat_names, ess_first_name_map


(
    essential_patterns,
    essential_flat_names,
    essentials_first_name_map
) = _render_essential_patterns(
    TAG_ESSENTIAL_FIELD_NAMES
)


def _this_extractor_a(_, __):
    return None


def _this_extractor_b(field_name, data_values):
    return data_values.get(field_name)


def _this_extractor_c(field_name, data_values):
    ptn = essential_patterns[field_name]
    return ptn.search(str(data_values)).group(1)


def do(data) -> Dict[str, any]:
    if data is None:
        return {}

    data_is_dict = isinstance(data, dict)
    data_field = {}
    for body_field_name in TAG_ESSENTIAL_BODY_NAMES:
        this_data_field = (
            data.get(body_field_name, '')
            if data_is_dict
            else getattr(data, body_field_name, '')
        )

        this_extractor = _this_extractor_a
        if isinstance(this_data_field, dict):
            this_extractor = _this_extractor_b
        elif (
            isinstance(this_data_field, str)
            or
            isinstance(this_data_field, bytes)
        ):
            this_extractor = _this_extractor_c

        data_field[body_field_name] = {
            'this_extractor': this_extractor,
            'this_data': this_data_field,
        }

    def find_value(field_name):
        result = None
        if data_is_dict and field_name in data:
            return data.get(field_name)

        for body_name in TAG_ESSENTIAL_BODY_NAMES:
            try:
                extractor = data_field[body_name]
                result = extractor['this_extractor'](
                    field_name,
                    extractor['this_data']
                )
                if result is not None:
                    break
            except (IndexError, AttributeError):
                continue

        return result

    found_values = dict(
        (
            essentials_first_name_map.get(field_name, field_name),
            found_value
        )
        for field_name, found_value in (
            (field_name, find_value(field_name))
            for field_name in essential_flat_names
        ) if found_value is not None
    )
    return found_values
